/*	$NetBSD: standout.c,v 1.19 2017/01/10 23:49:20 roy Exp $	*/

/*
 * Copyright (c) 1981, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <netbsd_sys/cdefs.h>

#include "curses.h"
#include "curses_private.h"

#ifndef _CURSES_USE_MACROS

/*
 * standout --
 *	Enter standout mode on stdscr.
 */
int
standout(void)
{

	return wstandout(stdscr);
}

/*
 * standend --
 *	Exit standout mode on stdscr.
 */
int
standend(void)
{

	return wstandend(stdscr);
}

#endif

/*
 * wstandout --
 *	Enter standout mode in window win.
 */
int
wstandout(WINDOW *win)
{
	const TERMINAL *t = win->screen->term;

	/*
	 * If standout/standend strings, or can underline, set the
	 * screen standout bit.
	 */
	if ((t_enter_standout_mode(t) != NULL &&
	    t_exit_standout_mode(t) != NULL) ||
	    t_underline_char(t) != NULL)
		win->wattr |= __STANDOUT;
	return 1;
}

/*
 * wstandend --
 *	Exit standout mode in window win.
 */
int
wstandend(WINDOW *win)
{

	// http://pubs.opengroup.org/onlinepubs/7908799/xcurses/wstandend.html
	win->wattr = __NORMAL;
	return 1;
}
